# -*- coding: utf-8 -*-
"""
@author: kebo
@contact: kebo0912@outlook.com
@version: 1.0

@file: __init__.py 
@time: 2020/6/23 下午11:55

这一行开始写关于本文件的说明与解释

"""


def func():
    pass


if __name__ == '__main__':
    func()
