# -*- coding: utf-8 -*-
"""
@author: kebo
@contact: kebo0912@outlook.com

@version: 1.0
@file: __init__.py
@time: 2019/11/25 17:24

这一行开始写关于本文件的说明与解释


"""
from leslie.predictors.text_classifier_predictor import TextBiClassifierPredictor


def func():
    pass


if __name__ == '__main__':
    func()
